# *** imports

# ** core
from typing import Dict, Any, Tuple

# ** app
from .request import RequestContext
from .feature import FeatureContext
from .error import ErrorContext
from .container import create_injector, import_dependency
from ..domain import Model, StringType, DictType, ModelType
from ..domain import AppInterface
from ..repos import AppRepository


# *** contexts

# ** context: app_context
class AppContext(Model):

    # * attribute: app_repo_module_path
    app_repo_module_path = StringType(
        required=True,
        default='tiferet.repos.app',
        metadata=dict(
            description='The application repository module path.'
        ),
    )

    # * attribute: app_repo_class_name
    app_repo_class_name = StringType(
        required=True,
        default='AppYamlProxy',
        metadata=dict(
            description='The application repository class name.'
        ),
    )

    # * attribute: app_repo_parameters
    app_repo_parameters = DictType(
        StringType(),
        default=dict(
            app_config_file='app/configs/app.yml'
        ),
        metadata=dict(
            description='The application repository parameters.'
        ),
    )

    # * method: run
    def run(self, interface_id: str, dependencies: Dict[str, Any] = {}, **kwargs) -> Any:
        '''
        Run the application interface.

        :param interface_id: The interface ID.
        :type interface_id: str
        :param dependencies: The dependencies.
        :type dependencies: dict
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The response.
        :rtype: Any
        '''

        # Load the interface.
        app_interface = self.load_interface(interface_id, **dependencies)

        # Run the interface.
        return app_interface.run(**kwargs)

    # * method: load_interface
    def load_interface(self, interface_id: str, **dependencies) -> AppInterface:
        '''
        Load the application interface.

        :param interface_id: The interface ID.
        :type interface_id: str
        :param dependencies: The dependencies.
        :type dependencies: dict
        :return: The application interface.
        :rtype: AppInterface
        '''
        
        # Import the app repository.
        app_repo = self.import_app_repo(
            self.app_repo_module_path, 
            self.app_repo_class_name, 
            **self.app_repo_parameters
        )

        # Get the app interface.
        app_interface = app_repo.get_interface(interface_id)

        # Create the injector.
        injector = self.create_injector(app_interface, **dependencies)

        # Load the app interface context.
        return getattr(injector, 'app_context')

    # * method: import_app_repo
    def import_app_repo(self, module_path: str, class_name: str, **kwargs) -> AppRepository:
        '''
        Import the app repository.

        :param module_path: The module path.
        :type module_path: str
        :param class_name: The class name.
        :type class_name: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The app repository.
        :rtype: AppRepository
        '''

        # Try to import the module provided.
        try:
            return import_dependency(module_path, class_name)(**kwargs)

        # Return None if nothing comes up.
        except:
            return None
        
    # ** method: create_injector
    def create_injector(self, app_interface: AppInterface, **kwargs) -> Any:
        '''
        Create the injector.

        :param app_interface: The app interface.
        :type app_interface: AppInterface
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The injector.
        :rtype: Any
        '''

        # Get the dependencies for the app interface.
        dependencies = dict(
            interface_id=app_interface.id,
            app_name=app_interface.name,
            feature_flag=app_interface.feature_flag,
            data_flag=app_interface.data_flag,
            app_context=import_dependency(
                **app_interface.get_dependency('app_context').to_primitive()
            ),
            **app_interface.constants
        )

        # Add the remaining dependencies from the app interface.
        dependencies.update({dep.attribute_id: import_dependency(dep.module_path, dep.class_name) for dep in app_interface.dependencies})

        # Create the injector.
        return create_injector(app_interface.id, **dependencies, **kwargs)


# ** context: app_interface_context
class AppInterfaceContext(Model): 
    '''
    The application interface context is a class that is used to create and run the application interface.
    '''

    # * attribute: interface_id
    interface_id = StringType(
        required=True,
        metadata=dict(
            description='The interface ID.'
        ),
    )

    # * attribute: name
    name = StringType(
        required=True,
        metadata=dict(
            description='The application name.'
        ),
    )

    # * field: features
    features = ModelType(
        FeatureContext,
        required=True,
        metadata=dict(
            description='The feature context.'
        ),
    )

    # * field: errors
    errors = ModelType(
        ErrorContext,
        required=True,
        metadata=dict(
            description='The error context.'
        ),
    )

    # * method: init
    def __init__(self, interface_id: str, app_name: str, feature_context: FeatureContext, error_context: ErrorContext):
        '''
        Initialize the application interface context.

        :param interface_id: The interface ID.
        :type interface_id: str
        :param app_name: The application name.
        :type app_name: str
        :param feature_context: The feature context.
        :type feature_context: FeatureContext
        :param error_context: The error context.
        :type error_context: ErrorContext
        '''

        # Initialize the model.
        super().__init__(dict(
            interface_id=interface_id,
            name=app_name
        ))
        self.features = feature_context
        self.errors = error_context

    # * method: parse_request
    def parse_request(self, request: Any, **kwargs) -> Tuple[RequestContext, dict]:
        '''
        Parse the incoming request.

        :param request: The incoming request.
        :type request: Any
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The request context.
        :rtype: RequestContext
        '''

        # Parse request.
        return request, kwargs
    
    # * method: execute_feature
    def execute_feature(self, request: RequestContext, **kwargs):
        '''
        Execute the feature context.

        :param request: The request context.
        :type request: RequestContext
        '''

        # Execute feature context and return session.
        self.features.execute(request, **kwargs)
    
    # * method: handle_response
    def handle_response(self, request: RequestContext) -> Any:
        '''
        Handle the response.

        :param request: The request context.
        :type request: RequestContext
        :return: The response.
        :rtype: Any
        '''
        
        # Import the JSON module.
        import json

        # Return the response.
        return json.loads(request.result) if request.result else None
    
    # * method: run
    def run(self, **kwargs):
        '''
        Run the application interface.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''
        
        # Parse request.
        request, kwargs = self.parse_request(**kwargs)

        # Execute feature context and return session.
        try:
            self.execute_feature(request, **kwargs)

        # Handle error and return response if triggered.
        except Exception as e:
            print('Error:', e)
            return self.errors.handle_error(e)

        # Handle response.
        return self.handle_response(request)



# *** imports

# ** app
from .app import AppContext, AppContext as App, AppInterfaceContext
from .container import ContainerContext
from .error import ErrorContext
from .feature import FeatureContext
from .request import RequestContext


from ..repos.container import ContainerRepository
from ..domain.container import ContainerAttribute, ContainerDependency
from ..services import container as container_service


class SetContainerAttribute(object):
    '''
    Command to set a new container attribute
    '''

    container_repo: ContainerRepository

    def __init__(self, container_repo: ContainerRepository):
        '''
        Initialize the command to set a new container attribute.

        :param container_repo: The container repository.
        :type container_repo: ContainerRepository
        '''

        self.container_repo = container_repo

    def execute(self, attribute_id: str, type: str, **kwargs):
        '''
        Execute the command to set a new container attribute.

        :param attribute_id: The attribute id.
        :type attribute_id: str
        :param type: The attribute type.
        :type type: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''
        
        # Look up the container attribute.
        attribute: ContainerAttribute = self.container_repo.get_attribute(attribute_id, type)

        # If not attribute is found, create a new one.
        if not attribute:
            attribute = ContainerAttribute.new(
                id=attribute_id,
                type=type,
                dependencies=[ContainerDependency.new(**kwargs)])
        
        # Otherwise, create the container depenedency and add it to the attribute.
        else:
            dependency = ContainerDependency.new(**kwargs)
            attribute.set_dependency(dependency)

        # Save the container attribute.
        self.container_repo.save_attribute(attribute=attribute)

        # Return the new container attribute.
        return attribute



# *** imports

# ** core
from typing import Dict

# ** app
from ..configs import *
from ..domain import DataObject
from ..domain.app import AppDependency, AppInterface


# *** constants

# ** constant: app_dependency_default
FEATURE_CONTEXT_DEFAULT = dict(
    module_path='tiferet.contexts.feature',
    class_name='FeatureContext',
) 

# * constant: app_dependency_default
CONTAINER_CONTEXT_DEFAULT = dict(
    module_path='tiferet.contexts.container',
    class_name='ContainerContext',
)

ERROR_CONTEXT_DEFAULT = dict(
    module_path='tiferet.contexts.error',
    class_name='ErrorContext',
)

FEATURE_REPO_DEFAULT = dict(
    module_path='tiferet.repos.feature',
    class_name='YamlProxy',
)

CONTAINER_REPO_DEFAULT = dict(
    module_path='tiferet.repos.container',
    class_name='YamlProxy',
)

ERROR_REPO_DEFAULT = dict(
    module_path='tiferet.repos.error',
    class_name='YamlProxy',
)

# ** constant: context_list_default
CONTEXT_LIST_DEFAULT = {
    'feature_context': FEATURE_CONTEXT_DEFAULT,
    'container_context': CONTAINER_CONTEXT_DEFAULT,
    'error_context': ERROR_CONTEXT_DEFAULT,
    'feature_repo': FEATURE_REPO_DEFAULT,
    'container_repo': CONTAINER_REPO_DEFAULT,
    'error_repo': ERROR_REPO_DEFAULT,
}

# *** data

# ** data: app_dependency_yaml_data
class AppDependencyYamlData(AppDependency, DataObject):
    '''
    A YAML data representation of an app dependency object.
    '''

    # * attribute: attribute_id
    attribute_id = StringType(
        metadata=dict(
            description='The attribute id for the application dependency.'
        ),
    )

    class Options():
        '''
        The options for the app dependency data.
        '''
        serialize_when_none = False
        roles = {
            'to_model': DataObject.allow(),
            'to_data.yaml': DataObject.deny('attribute_id')
        }

    # * method: from_data
    @staticmethod
    def from_data(**kwargs) -> 'AppDependencyYamlData':
        '''
        Initializes a new YAML representation of an AppDependency object.
        
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new AppDependencyData object.
        :rtype: AppDependencyData
        '''

        # Create a new AppDependencyData object.
        return super(AppDependencyYamlData, AppDependencyYamlData).from_data(
            AppDependencyYamlData,
            **kwargs
        )

    # * method: new
    @staticmethod
    def from_data(**kwargs) -> 'AppDependencyYamlData':
        '''
        Initializes a new YAML representation of an AppDependency object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new AppDependencyData object.
        :rtype: AppDependencyData
        '''

        # Create a new AppDependencyData object.
        return super(AppDependencyYamlData, AppDependencyYamlData).from_data(
            AppDependencyYamlData,
            **kwargs
        )

    # * method: map
    def map(self, **kwargs) -> AppDependency:
        '''
        Maps the app dependency data to an app dependency object.

        :param role: The role for the mapping.
        :type role: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new app dependency object.
        :rtype: AppDependency
        '''

        # Map the app dependency data.
        return super().map(AppDependency, **kwargs)


# ** data: app_interface_yaml_data
class AppInterfaceYamlData(AppInterface, DataObject):
    '''
    A data representation of an app interface object.
    '''

    class Options():
        '''
        The options for the app interface data.
        '''
        serialize_when_none = False
        roles = {
            'to_model': DataObject.deny('app_context', 'container_context', 'feature_context', 'error_context', 'feature_repo', 'container_repo', 'error_repo'),
            'to_data': DataObject.deny('id')
        }

    # attribute: app_context
    app_context = ModelType(
        AppDependencyYamlData,
        required=True,
        metadata=dict(
            description='The application context dependency.'
        ),
    )

    # * attribute: feature_context
    feature_context = ModelType(
        AppDependencyYamlData,
        required=True,
        metadata=dict(
            description='The feature context dependency.'
        ),
    )

    # * attribute: container_context
    container_context = ModelType(
        AppDependencyYamlData,
        required=True,
        metadata=dict(
            description='The container context dependency.'
        ),
    )

    # * attribute: error_context
    error_context = ModelType(
        AppDependencyYamlData,
        required=True,
        metadata=dict(
            description='The error context dependency.'
        ),
    )

    # * attribute: feature_repo
    feature_repo = ModelType(
        AppDependencyYamlData,
        required=True,
        metadata=dict(
            description='The feature repository dependency.'
        ),
    )

    # * attribute: container_repo
    container_repo = ModelType(
        AppDependencyYamlData,
        required=True,
        metadata=dict(
            description='The container repository dependency.'
        ),
    )

    # * attribute: error_repo
    error_repo = ModelType(
        AppDependencyYamlData,
        required=True,
        metadata=dict(
            description='The error repository dependency.'
        ),
    )

    # * method: new
    @staticmethod
    def from_data(app_context: Dict[str, str],
        **kwargs) -> 'AppInterfaceYamlData':
        '''
        Initializes a new YAML representation of an AppInterface object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new AppInterfaceData object.
        :rtype: AppInterfaceData
        '''

        # Add the app context to the dependencies.
        dependencies = dict(
            app_context=AppDependencyYamlData.from_data(
                attribute_id='app_context',
                **app_context
            )
        )

        # Going through the default dependencies...
        for key, value in CONTEXT_LIST_DEFAULT.items():
            
            # If the key is in the kwargs, add it and continue.
            if key in kwargs:
                dependencies[key] = AppDependencyYamlData.from_data(
                    attribute_id=key,
                    **kwargs.pop(key)) # Pop the key to avoid duplication.
                continue
            
            # Otherwise, add the default value.
            dependencies[key] = AppDependencyYamlData.from_data(
                attribute_id=key,
                **value)

        # Create a new AppInterfaceData object.
        return super(AppInterfaceYamlData, AppInterfaceYamlData).from_data(
            AppInterfaceYamlData,
            **dependencies,
            **kwargs
        )

    # * method: map
    def map(self, **kwargs) -> AppInterface:
        '''
        Maps the app interface data to an app interface object.

        :param role: The role for the mapping.
        :type role: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new app interface object.
        :rtype: AppInterface
        '''

        # Format and map the dependencies.
        dependencies = [
            self.app_context.map(),
            self.container_context.map(),
            self.feature_context.map(),
            self.error_context.map(),
            self.feature_repo.map(),
            self.container_repo.map(),
            self.error_repo.map(),
        ]

        # Map the app interface data.
        return super().map(AppInterface,
            dependencies=dependencies,
            **self.to_primitive('to_model'),
            **kwargs
        )



# *** imports

# ** core
import os
from typing import Any

# ** app
from ..domain import *
from ..repos.container import ContainerRepository


# *** functions

# ** function: create_injector
def create_injector(name: str, **dependencies) -> Any:
    '''
    Create an injector object with the given dependencies.

    :param name: The name of the injector.
    :type name: str
    :param dependencies: The dependencies.
    :type dependencies: dict
    :return: The injector object.
    :rtype: Any
    '''

    # Create container.
    from dependencies import Injector
    return type(f'{name.capitalize()}Container', (Injector,), {**dependencies})


# ** function: import_dependency
def import_dependency(module_path: str, class_name: str) -> Any:
    '''
    Import an object dependency from its configured Python module.

    :param module_path: The module path.
    :type module_path: str
    :param class_name: The class name.
    :type class_name: str
    :return: The dependency.
    :rtype: Any
    '''

    # Import module.
    from importlib import import_module
    return getattr(import_module(module_path), class_name)


# *** contexts

# ** contexts: container_context
class ContainerContext(Model):
    '''
    A container context is a class that is used to create a container object.
    '''

    # * attribute: interface_id
    interface_id = StringType(
        required=True,
        metadata=dict(
            description='The interface ID.'
        ),
    )

    # * attribute: attributes
    attributes = DictType(
        ModelType(ContainerAttribute), 
        default={}, 
        required=True,
        metadata=dict(
            description='The container attributes.'
        ),
    )

    # * attribute: constants
    constants = DictType(
        StringType, 
        default={},
        metadata=dict(
            description='The container constants.'
        ),
    )

    # * attribute: feature_flag
    feature_flag = StringType(
        required=True,
        default='core',
        metadata=dict(
            description='The feature flag.'
        ),
    )

    # * attribute: data_flag
    data_flag = StringType(
        required=True,
        metadata=dict(
            description='The data flag.'
        ),
    )

    # * method: init
    def __init__(self, interface_id: str, container_repo: ContainerRepository, feature_flag: str, data_flag: str):
        '''
        Initialize the container context.

        :param interface_id: The interface ID.
        :type interface_id: str
        :param container_repo: The container repository.
        :type container_repo: ContainerRepository
        :param interface_flag: The interface flag.
        :type interface_flag: str
        :param feature_flag: The feature flag.
        :type feature_flag: str 
        :param data_flag: The data flag.
        :type data_flag: str
        '''

        # Add the attributes and constants as empty dictionaries.
        attributes = {}
        constants = {}
        
        # Get and set attributes and constants.
        attrs, consts = container_repo.list_all()

        # Parse the constants.
        constants.update({key: self.parse_parameter(consts[key]) for key in consts})
        
        # Add the attributes to the context.
        for attr in attrs:
            
            # If the attribute already exists, set the dependencies.
            if attr.id in attributes:
                for dep in attr.dependencies:
                    attr.set_dependency(dep)
                    continue

            # Otherwise, add the attribute.
            attributes[attr.id] = attr

            # Add any parameters as constants.
            for dep in attr.dependencies:
                for key in dep.parameters:
                    constants[key] = self.parse_parameter(dep.parameters[key])

        # Add the constants and attributes to the context.
        super().__init__(dict(
            interface_id=interface_id,
            feature_flag=feature_flag,
            data_flag=data_flag,
            attributes=attributes,
            constants=constants,
        ))

    
    # * method: parse_environment_parameter
    def parse_parameter(self, parameter: str) -> str:

        # If the parameter is an environment variable, get the value.
        if parameter.startswith('$env.'):
            return os.getenv(parameter[5:])
        
        # Otherwise, return the parameter.
        return parameter

    # * method: get_dependency
    def get_dependency(self, attribute_id: str):
        '''
        Get a dependency from the container.

        :param attribute_id: The attribute id of the dependency.
        :type attribute_id: str
        :return: The attribute value.
        :rtype: Any
        '''

        # Create the injector.
        injector = self.create_injector()

        # Get attribute.
        return getattr(injector, attribute_id)

    # * method: create_injector
    def create_injector(self, **kwargs) -> Any:
        '''
        Add a container to the context.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The container injector object.
        :rtype: Any
        '''

        # Import dependencies.
        dependencies = {}
        for attribute_id in self.attributes:
            attribute = self.attributes[attribute_id]
            flag_map = dict(
                feature=self.feature_flag,
                data=self.data_flag,
            )
            dependencies[attribute_id] = self.import_dependency(attribute, flag_map[attribute.type])

        # Create container.
        return create_injector( 
            self.interface_id, 
            **self.constants, 
            **dependencies, 
            **kwargs)

    # * method: import_dependency
    def import_dependency(self, attribute: ContainerAttribute, flag: str) -> Any:
        '''
        Import a container attribute dependency from its configured Python module.

        :param attribute: The container attribute.
        :type attribute: ContainerAttribute
        :param flag: The flag for the dependency.
        :type flag: str
        :return: The dependency.
        :rtype: Any
        '''

        # Get the dependency.
        dependency = attribute.get_dependency(flag)

        # If there is no dependency and the attribute is a feature, import the default feature.
        if not dependency and attribute.type == 'feature':
            dependency = attribute.get_dependency('core')

        # Import the dependency.
        return import_dependency(dependency.module_path, dependency.class_name)


# *** imports

# ** app
from .container import ContainerContext
from .request import RequestContext
from ..domain import *
from ..repos.feature import FeatureRepository


# *** contexts

# ** context: feature_context
class FeatureContext(Model):

    # * attribute: features
    features = DictType(
        ModelType(Feature),
        required=True,
        metadata=dict(
            description='The features lookup.'
        )
    )

    # * attribute: container
    container = ModelType(
        ContainerContext,
        required=True,
        metadata=dict(
            description='The container context.'
        ),
    )

    # * method: init
    def __init__(self, feature_repo: FeatureRepository, container_context: ContainerContext):
        '''
        Initialize the feature context.

        :param feature_repo: The feature repository.
        :type feature_repo: FeatureRepository
        :param container_context: The container context.
        :type container_context: ContainerContext
        '''

        # Create the features.
        features = {feature.id: feature for feature in feature_repo.list()}

        # Set the features and container.
        ## NOTE: There is a bug in the schematics library that does not allow us to initialize 
        ## the feature context with the container context directly.
        super().__init__(dict(
            features=features,
        ))
        self.container = container_context

    # * method: parse_parameter
    def parse_parameter(self, parameter: str) -> str:
        '''
        Parse a parameter.

        :param parameter: The parameter to parse.
        :type parameter: str
        :return: The parsed parameter.
        :rtype: str
        '''

        # Parse the parameter.
        return self.container.parse_parameter(parameter)
        
    # * method: execute
    def execute(self, request: RequestContext, debug: bool = False, **kwargs):
        '''
        Execute the feature request.
        
        :param request: The request context object.
        :type request: r.RequestContext
        :param debug: Debug flag.
        :type debug: bool
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''

        # Assert the feature exists.
        assert request.feature_id in self.features, 'FEATURE_NOT_FOUND, {}'.format(request.feature_id)

        # Iterate over the feature commands.
        for command in self.features[request.feature_id].commands:

            # Get the feature command handler instance.
            handler = self.container.get_dependency(command.attribute_id)

            # Parse the command parameters
            params = {
                param: 
                self.parse_parameter(
                    command.params.get(param)
                ) 
                for param in command.params
            }

            # Execute the handler function.
            # Handle assertion errors if pass on error is not set.
            try:
                result = handler.execute(
                    **request.data,
                    **params,
                    debug=debug,
                    **kwargs)
                
                # Return the result to the session context if return to data is set.
                if command.return_to_data:
                    request.data[command.data_key] = result
                    continue

                # Set the result in the request context.
                if result:
                    request.set_result(result)

            # Handle assertion errors if pass on error is not set.
            except AssertionError as e:
                if not command.pass_on_error:
                    raise e 

            



from ..domain.error import Error
from ..repos.error import ErrorRepository

class AddNewError(object):

    def __init__(self, error_repo: ErrorRepository):
        self.error_repo = error_repo

    def execute(self, **kwargs) -> Error:

        # Create a new error.
        error: Error = Error.new(**kwargs)

        # Assert that the error does not already exist.
        assert not self.error_repo.exists(error.id), f'ERROR_ALREADY_EXISTS: {error.id}'

        # Save the error.
        self.error_repo.save(error)

        # Return the new error.
        return error